import { AlchemyProvider, StaticJsonRpcProvider } from '@ethersproject/providers';
import { ABI_Item, ABI_ItemUnfiltered } from './interfaces/abi';
import { InterpreterMap } from './interfaces/contractInterpreter';
import { ContractData, ContractType, DecodedCallData, DecodedTx, Interaction } from './interfaces/decoded';
import { ActivityData, Interpretation } from './interfaces/interpreted';
import { RawTxData } from './interfaces/rawData';
import { EVMTransaction } from './interfaces/s3';
import { Chain } from './interfaces/utils';
import { ZenLedgerData, ZenLedgerRow } from './interfaces/zenLedger';
import Covalent from './utils/clients/Covalent';
import Etherscan, { EtherscanServiceLevel } from './utils/clients/Etherscan';
import { DatabaseInterface } from './utils/DatabaseInterface';
import { Augmenter } from './core/Augmenter';
import Interpreter from './core/Interpreter';
import RawDataFetcher from './core/RawDataFetcher';
export declare type TranslatorConfig = {
    chain: Chain;
    nodeUrl?: string;
    alchemyProjectId: string;
    etherscanAPIKey: string;
    userAddress?: string;
    connectionString?: string;
    etherscanServiceLevel?: EtherscanServiceLevel;
    covalentAPIKey?: string;
    additionalInterpreters?: Record<string, InterpreterMap>;
};
export declare type NamesAndSymbolsMap = Record<string, {
    name: string | null;
    symbol: string | null;
}>;
declare class Translator {
    nodeUrl: string | null;
    alchemyProjectId: string;
    chain: Chain;
    provider: AlchemyProvider | StaticJsonRpcProvider;
    rawDataFetcher: RawDataFetcher;
    etherscan: Etherscan;
    userAddress: string | null;
    augmenter: Augmenter;
    interpreter: Interpreter;
    databaseInterface: DatabaseInterface;
    covalent: Covalent | null;
    constructor(config: TranslatorConfig);
    initializeMongoose(): Promise<void>;
    private getProvider;
    updateUserAddress(userAddress: string): void;
    /**********************************************/
    /**** FROM WALLET ADDRESS OR TX HASH ONLY *****/
    /**********************************************/
    getTxHashesByBlockNumber(blockNumber: string): Promise<string[]>;
    getRawTxData(txHash: string): Promise<RawTxData>;
    getRawTxDataWithoutTrace(txHash: string): Promise<RawTxData>;
    getTxHashArrayForAddress(address: string, limit?: number): Promise<string[]>;
    getRawDataFromS3Data(tx: EVMTransaction, timestamp: number): RawTxData;
    getContractAddressesFromRawTxData(rawTxData: RawTxData): string[];
    /**********************************************/
    /**** USING TX AND CONTRACT ADDRESSES  ********/
    /****  Data used to decode & augment   ********/
    /**********************************************/
    getAllAddresses(decodedLogs: Interaction[], decodedCallData: DecodedCallData, contractAddresses: string[]): string[];
    getABIsAndNamesForContracts(contractAddresses: string[]): Promise<[Record<string, ABI_ItemUnfiltered[]>, Record<string, string | null>]>;
    getProxyContractMap(contractAddresses: string[]): Promise<Record<string, string>>;
    getNameAndSymbol(address: string, contractType: ContractType): Promise<{
        tokenName: string | null;
        tokenSymbol: string | null;
        contractName: string | null;
    }>;
    downloadContractsFromTinTin(): Promise<ContractData[]>;
    getENSName(address: string): Promise<string>;
    getENSNames(addresses: string[]): Promise<Record<string, string>>;
    getContractsData(contractToAbiMap: Record<string, ABI_ItemUnfiltered[]>, contractToOfficialNameMap: Record<string, string | null>, proxyAddressMap: Record<string, string>): Promise<Record<string, ContractData>>;
    getContractType(address: string, abi: ABI_Item[]): Promise<ContractType>;
    /**********************************************/
    /******      DECODING / AUGMENTING     ********/
    /**********************************************/
    decodeTxData(rawTxData: RawTxData, ABIs: Record<string, ABI_Item[]>, contractDataMap: Record<string, ContractData>): Promise<{
        decodedLogs: Interaction[];
        decodedCallData: DecodedCallData;
        decodedTraceData: DecodedCallData[];
    }>;
    decodeTxDataArr(rawTxDataArr: RawTxData[], ABIs: Record<string, ABI_Item[]>[]): DecodedTx[];
    augmentDecodedData(decodedLogs: Interaction[], decodedCallData: DecodedCallData, decodedTraceData: DecodedCallData[], ensMap: Record<string, string>, contractDataMap: Record<string, ContractData>, rawTxData: RawTxData): DecodedTx;
    /**********************************************/
    /******          INTERPRETING          ********/
    /**********************************************/
    interpretDecodedTx(decoded: DecodedTx, userAddress?: string | null, userName?: string | null, entityMap?: Record<string, string | null>): Promise<Interpretation>;
    interpretDecodedTxArr(decodedTxArr: (DecodedTx | null)[], userAddress?: string | null, userName?: string | null): Promise<(Interpretation | null)[]>;
    /**********************************************/
    /******            COMBINED            ********/
    /**********************************************/
    interpretTx(txHash: string, userAddress?: string | null): Promise<Interpretation>;
    decodeFromTxHash(txHash: string, excludeTraceLogs?: boolean): Promise<{
        decodedTx: DecodedTx;
        rawTxData: RawTxData;
    }>;
    decodeFromTxHashArr(txHashArr: string[], excludeTraceLogs?: boolean): Promise<DecodedTx[]>;
    allDataFromTxHash(txHash: string, providedUserAddress?: string | null): Promise<ActivityData>;
    allDataFromTxHashArr(txHashArr: string[], userAddressUnvalidated?: string | null): Promise<ActivityData[]>;
    /**********************************************/
    /******       DB ONLY CALLs            ********/
    /**********************************************/
    getManyDecodedTxFromDB(txHashArr: string[]): Promise<(DecodedTx | null)[]>;
    translateWithTaxData(zenLedgerData: ZenLedgerData[], address: string): Promise<ZenLedgerRow[]>;
}
export default Translator;
