"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const alchemy_sdk_1 = require("@alch/alchemy-sdk");
const providers_1 = require("@ethersproject/providers");
const utils_1 = require("./interfaces/utils");
const utils_2 = require("./utils");
const Covalent_1 = __importDefault(require("./utils/clients/Covalent"));
const Etherscan_1 = __importDefault(require("./utils/clients/Etherscan"));
const DatabaseInterface_1 = require("./utils/DatabaseInterface");
const logging_1 = require("./utils/logging");
const mongoose_1 = require("./utils/mongoose");
const timer_1 = __importDefault(require("./utils/timer"));
const Augmenter_1 = require("./core/Augmenter");
const Interpreter_1 = __importDefault(require("./core/Interpreter"));
const RawDataFetcher_1 = __importDefault(require("./core/RawDataFetcher"));
const TaxFormatter_1 = __importDefault(require("./core/TaxFormatter"));
// const defaultConfig: TranslatorConfigTwo = {
//     chain: getChainBySymbol('ETH'),
// }
class Translator {
    constructor(config) {
        this.chain = config.chain;
        this.nodeUrl = config.nodeUrl || null;
        this.alchemyProjectId = config.alchemyProjectId;
        this.userAddress = config.userAddress ? utils_1.AddressZ.parse(config.userAddress) : null;
        this.provider = this.getProvider();
        this.etherscan = new Etherscan_1.default(config.etherscanAPIKey, config.etherscanServiceLevel);
        this.databaseInterface = config.connectionString
            ? new mongoose_1.MongooseDatabaseInterface(config.connectionString)
            : new DatabaseInterface_1.NullDatabaseInterface();
        this.rawDataFetcher = new RawDataFetcher_1.default(this.provider);
        this.augmenter = new Augmenter_1.Augmenter(this.provider, null, this.etherscan, this.databaseInterface);
        this.interpreter = new Interpreter_1.default(config.chain, null, this.databaseInterface, config.additionalInterpreters);
        this.covalent = config.covalentAPIKey ? new Covalent_1.default(config.covalentAPIKey, config.chain.id) : null;
    }
    initializeMongoose() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.databaseInterface instanceof mongoose_1.MongooseDatabaseInterface) {
                yield this.databaseInterface.connect();
            }
        });
    }
    getProvider() {
        if (this.nodeUrl) {
            return new providers_1.StaticJsonRpcProvider(this.nodeUrl, this.chain.id);
        }
        if (this.alchemyProjectId) {
            const settings = {
                apiKey: this.alchemyProjectId,
                network: this.chain.alchemyNetworkString,
                maxRetries: 10,
            };
            const alchemy = (0, alchemy_sdk_1.initializeAlchemy)(settings);
            const provider = alchemy.getProvider();
            return provider;
        }
        else {
            throw new Error('No node url or alchemy project id provided');
        }
    }
    updateUserAddress(userAddress) {
        this.userAddress = utils_1.AddressZ.parse(userAddress);
    }
    /**********************************************/
    /**** FROM WALLET ADDRESS OR TX HASH ONLY *****/
    /**********************************************/
    getTxHashesByBlockNumber(blockNumber) {
        return __awaiter(this, void 0, void 0, function* () {
            return this, this.rawDataFetcher.getTxHashesByBlockNumber(blockNumber);
        });
    }
    // Gets the txResponse, txReceipt, txTrace from a node. Archive node needed
    getRawTxData(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.rawDataFetcher.getTxData(txHash);
        });
    }
    getRawTxDataWithoutTrace(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.rawDataFetcher.getTxDataWithoutTrace(txHash);
        });
    }
    getTxHashArrayForAddress(address, 
    // includeInitiatedTxs = true,
    // includeNotInitiatedTxs = false,
    limit = 100) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.covalent)
                throw new Error('Covalent not configured');
            const ens = (yield this.getENSNames([address]))[address];
            timer_1.default.startTimer(`covalent for ${ens || address}`);
            const allCovalentTxDataArr = yield this.covalent.getTransactionsFor(utils_1.AddressZ.parse(address), limit);
            timer_1.default.stopTimer(`covalent for ${ens || address}`);
            return allCovalentTxDataArr.map((txData) => txData.tx_hash);
        });
    }
    getRawDataFromS3Data(tx, timestamp) {
        return this.rawDataFetcher.getTxDataFromS3Tx(tx, timestamp);
    }
    // Could rely on Etherscan, but they have a max of 10k records
    // async getAllTxHashesForAddress(address: string): Promise<string[]> {
    //     throw new Error('Not implemented')
    // }
    getContractAddressesFromRawTxData(rawTxData) {
        return RawDataFetcher_1.default.getContractAddressesFromRawTxData(rawTxData);
    }
    /**********************************************/
    /**** USING TX AND CONTRACT ADDRESSES  ********/
    /****  Data used to decode & augment   ********/
    /**********************************************/
    getAllAddresses(decodedLogs, decodedCallData, contractAddresses) {
        return Augmenter_1.Augmenter.getAllAddresses(decodedLogs, decodedCallData, contractAddresses);
    }
    getABIsAndNamesForContracts(contractAddresses) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.getABIsAndNamesForContracts(contractAddresses);
        });
    }
    getProxyContractMap(contractAddresses) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.getProxyContractMap(contractAddresses);
        });
    }
    getNameAndSymbol(address, contractType) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.getNameAndSymbol(address, contractType);
        });
    }
    downloadContractsFromTinTin() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.downloadContractsFromTinTin();
        });
    }
    getENSName(address) {
        return __awaiter(this, void 0, void 0, function* () {
            const nameMap = yield this.augmenter.getENSNames([address]);
            return nameMap[address];
        });
    }
    getENSNames(addresses) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.getENSNames(addresses);
        });
    }
    getContractsData(contractToAbiMap, contractToOfficialNameMap, proxyAddressMap) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.augmenter.getContractsData(contractToAbiMap, contractToOfficialNameMap, proxyAddressMap);
        });
    }
    // We should store this in a contracts table just so we can see this data more easily
    getContractType(address, abi) {
        return this.augmenter.getContractType(address, abi);
    }
    /**********************************************/
    /******      DECODING / AUGMENTING     ********/
    /**********************************************/
    decodeTxData(rawTxData, ABIs, contractDataMap) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.augmenter.decodeTxData(rawTxData, ABIs, contractDataMap);
        });
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    decodeTxDataArr(rawTxDataArr, ABIs) {
        throw new Error('Not implemented');
    }
    augmentDecodedData(decodedLogs, decodedCallData, decodedTraceData, ensMap, contractDataMap, rawTxData) {
        return this.augmenter.augmentDecodedData(decodedLogs, decodedCallData, decodedTraceData, ensMap, contractDataMap, rawTxData);
    }
    /**********************************************/
    /******          INTERPRETING          ********/
    /**********************************************/
    interpretDecodedTx(decoded, userAddress = null, userName = null, entityMap = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.interpreter.interpretSingleTx(decoded, userAddress, userName, entityMap);
        });
    }
    interpretDecodedTxArr(decodedTxArr, userAddress = null, userName = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const addresses = [
                ...new Set(decodedTxArr.map((decoded) => [decoded === null || decoded === void 0 ? void 0 : decoded.toAddress, decoded === null || decoded === void 0 ? void 0 : decoded.fromAddress]).flat()),
            ].filter((str) => str !== null && str !== undefined);
            if (userAddress)
                addresses.push(userAddress);
            const entityMap = yield this.databaseInterface.getManyEntityMap(addresses);
            return Promise.all(decodedTxArr.map((decodedTx) => __awaiter(this, void 0, void 0, function* () {
                return decodedTx ? yield this.interpretDecodedTx(decodedTx, userAddress, userName) : null;
            })));
        });
    }
    // If we do this after we've created the example description, we'll have to figure out how to parse any addresses we've turned into a shorter name or onoma name
    // addEnsToInterpretedData(ens: Record<string, string>): Interpretation {
    //     throw new Error('Not implemented')
    // }
    /**********************************************/
    /******            COMBINED            ********/
    /**********************************************/
    interpretTx(txHash, userAddress = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const rawTxData = yield this.getRawTxData(txHash);
            const addresses = this.getContractAddressesFromRawTxData(rawTxData);
            const [unfilteredAbiMap, officialContractNamesMap] = yield this.getABIsAndNamesForContracts(addresses);
            const proxyAddressMap = yield this.getProxyContractMap(addresses);
            const AbiMap = (0, utils_2.filterABIMap)(unfilteredAbiMap);
            const ensMap = yield this.getENSNames(addresses);
            const contractDataMap = yield this.getContractsData(AbiMap, officialContractNamesMap, proxyAddressMap);
            const { decodedLogs, decodedCallData, decodedTraceData } = yield this.decodeTxData(rawTxData, AbiMap, contractDataMap);
            const decodedWithAugmentation = this.augmentDecodedData(decodedLogs, decodedCallData, decodedTraceData, ensMap, contractDataMap, rawTxData);
            const interpretation = this.interpretDecodedTx(decodedWithAugmentation, userAddress);
            return interpretation;
        });
    }
    decodeFromTxHash(txHash, excludeTraceLogs = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const logData = {
                tx_hash: txHash,
            };
            try {
                logData.function_name = 'getRawTxData';
                const rawTxData = excludeTraceLogs
                    ? yield this.getRawTxDataWithoutTrace(txHash)
                    : yield this.getRawTxData(txHash);
                logData.address = rawTxData.txResponse.from;
                const contractAddresses = this.getContractAddressesFromRawTxData(rawTxData);
                // const proxyAddressMap = await this.getProxyContractMap(contractAddresses)
                // console.log('proxyAddressMap', proxyAddressMap)
                const proxyAddressMap = {};
                const contractAndProxyAddresses = [...contractAddresses, ...(0, utils_2.getValues)(proxyAddressMap)];
                logData.function_name = 'getABIsAndNamesForContracts';
                const [unfilteredAbiMap, officialContractNamesMap] = yield this.getABIsAndNamesForContracts(contractAndProxyAddresses);
                logData.function_name = 'getContractsData';
                const contractDataMap = yield this.getContractsData(unfilteredAbiMap, officialContractNamesMap, proxyAddressMap);
                // contractDataMap = await this.augmentProxyContractABIs(contractDataMap)
                const AbiMap = (0, utils_2.filterABIMap)(unfilteredAbiMap);
                logData.function_name = 'decodeTxData';
                const { decodedLogs, decodedCallData, decodedTraceData } = yield this.decodeTxData(rawTxData, AbiMap, contractDataMap);
                const allAddresses = this.getAllAddresses(decodedLogs, decodedCallData, contractAddresses);
                logData.function_name = 'getENSNames';
                const ensMap = yield this.getENSNames(allAddresses);
                const decodedWithAugmentation = this.augmentDecodedData(decodedLogs, decodedCallData, decodedTraceData, ensMap, contractDataMap, rawTxData);
                return { decodedTx: decodedWithAugmentation, rawTxData };
            }
            catch (error) {
                (0, logging_1.logError)(logData, error);
                throw error;
            }
        });
    }
    decodeFromTxHashArr(txHashArr, excludeTraceLogs = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = txHashArr.map((txHash) => {
                return this.decodeFromTxHash(txHash, excludeTraceLogs);
            });
            const results = yield Promise.all(promises);
            const decodedTxArr = results.map((result) => result.decodedTx);
            yield this.initializeMongoose();
            yield this.databaseInterface.addOrUpdateManyDecodedTx(decodedTxArr);
            return decodedTxArr;
        });
    }
    allDataFromTxHash(txHash, providedUserAddress = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const logData = {
                tx_hash: txHash,
            };
            try {
                logData.function_name = 'decodeFromTxHash';
                const { decodedTx, rawTxData } = yield this.decodeFromTxHash(txHash);
                const userAddressUnvalidated = providedUserAddress || decodedTx.fromAddress;
                const userAddress = utils_1.AddressZ.parse(userAddressUnvalidated);
                logData.address = userAddress;
                logData.function_name = 'getENSName';
                const userName = yield this.getENSName(userAddress);
                logData.function_name = 'interpretDecodedTx';
                const interpretation = yield this.interpretDecodedTx(decodedTx, userAddress, userName);
                return { interpretedData: interpretation, decodedTx, rawTxData };
            }
            catch (error) {
                (0, logging_1.logError)(logData, error);
                throw error;
            }
        });
    }
    allDataFromTxHashArr(txHashArr, userAddressUnvalidated = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = txHashArr.map((txHash) => {
                return this.allDataFromTxHash(txHash, userAddressUnvalidated);
            });
            const results = yield Promise.all(promises);
            const decodedTxArr = results.map((result) => result.decodedTx);
            yield this.initializeMongoose();
            yield this.databaseInterface.addOrUpdateManyDecodedTx(decodedTxArr);
            return results;
        });
    }
    /**********************************************/
    /******       DB ONLY CALLs            ********/
    /**********************************************/
    getManyDecodedTxFromDB(txHashArr) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(this.databaseInterface instanceof mongoose_1.MongooseDatabaseInterface)) {
                throw new Error('This function only works with MongodB');
            }
            yield this.initializeMongoose();
            const map = yield this.databaseInterface.getManyDecodedTxMap(txHashArr);
            return (0, utils_2.getValues)(map);
        });
    }
    translateWithTaxData(zenLedgerData, address) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const taxFormatter = new TaxFormatter_1.default(address, 'brenners wallet', this.chain);
                const rows = taxFormatter.format(zenLedgerData);
                zenLedgerData.forEach((element, index) => {
                    ;
                    element.taxData = rows[index];
                });
                return rows;
            }
            catch (error) {
                console.log('error in translateWithTaxData', error);
                return [];
            }
        });
    }
}
exports.default = Translator;
//# sourceMappingURL=Translator.js.map