"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable @typescript-eslint/no-non-null-assertion */
const contractInterpreters_1 = __importDefault(require("./contractInterpreters"));
const DoubleSidedTxInterpreter_1 = require("./DoubleSidedTxInterpreter");
const ContractDeploy_json_1 = __importDefault(require("./genericInterpreters/ContractDeploy.json"));
const gnosis_1 = __importDefault(require("./genericInterpreters/gnosis"));
const lastFallback_1 = __importDefault(require("./genericInterpreters/lastFallback"));
const token_1 = __importDefault(require("./genericInterpreters/token"));
const transfer_1 = __importDefault(require("./genericInterpreters/transfer"));
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const onoma_1 = require("onoma");
const utils_2 = require("../interfaces/utils");
const utils_3 = require("../utils");
const constants_1 = require("../utils/constants");
const DatabaseInterface_1 = require("../utils/DatabaseInterface");
function deepCopy(obj) {
    return JSON.parse(JSON.stringify(obj));
}
function includes(collection, element) {
    return collection.includes(element);
}
const topLevelInteractionKeys = ['contractName', 'contractSymbol', 'contractAddress'];
class Interpreter {
    constructor(chain, userAddress = null, db = null, additionalContractInterpreters) {
        this.contractSpecificInterpreters = {};
        this.chain = chain;
        this.userAddress = (userAddress && utils_2.AddressZ.parse(userAddress)) || null;
        this.db = db || new DatabaseInterface_1.NullDatabaseInterface();
        for (const [address, map] of Object.entries(contractInterpreters_1.default)) {
            this.contractSpecificInterpreters[address.toLowerCase()] = map;
        }
        if (additionalContractInterpreters) {
            for (const [address, map] of Object.entries(additionalContractInterpreters)) {
                this.contractSpecificInterpreters[address.toLowerCase()] = map;
            }
        }
    }
    updateUserAddress(userAddress) {
        this.userAddress = utils_2.AddressZ.parse(userAddress);
    }
    updateChain(chain) {
        this.chain = chain;
    }
    interpretSingleTx(decodedTx, userAddressFromInput = null, userNameFromInput = null, entityMap = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            // Prep data coming in from 'decodedData'
            const { methodCall: { name: methodName }, traceCalls, interactions, fromAddress, toAddress, timestamp, nativeValueSent, txHash, contractName, officialContractName, } = decodedTx;
            const gasUsed = (0, utils_1.formatEther)(ethers_1.BigNumber.from(decodedTx.gasUsed).mul(ethers_1.BigNumber.from(decodedTx.effectiveGasPrice)));
            const parsedAddress = utils_2.AddressZ.safeParse(userAddressFromInput);
            const userAddress = parsedAddress.success ? parsedAddress.data : this.userAddress || fromAddress;
            let userName = null;
            if (fromAddress === userAddress) {
                userName = decodedTx.fromENS;
            }
            if (toAddress === userAddress) {
                userName = decodedTx.toENS;
            }
            const addressesToCheck = [fromAddress, userAddress];
            if (toAddress)
                addressesToCheck.push(toAddress);
            // timer.startTimer(txHash.substring(2, 8))
            if ((0, utils_3.getKeys)(entityMap).length === 0) {
                entityMap = yield this.db.getManyEntityMap(addressesToCheck);
            }
            // timer.stopTimer(txHash.substring(2, 8))
            userName =
                userName ||
                    userNameFromInput ||
                    entityMap[userAddress] ||
                    (0, onoma_1.addressToName)(userAddress) ||
                    userAddress.substring(0, 6);
            // TODO generalize this so it'll get any ENS (ex: _operatorENS)
            // not contract-specific
            const interpretation = {
                txHash,
                userAddress,
                contractAddress: toAddress,
                actions: [],
                assetsReceived: this.getAssetsReceived(interactions, userAddress),
                assetsSent: this.getAssetsSent(interactions, nativeValueSent, userAddress, fromAddress),
                chainSymbol: this.chain.symbol,
                userName,
                gasPaid: gasUsed,
                extra: {},
                exampleDescription: 'no example description defined',
                reverted: decodedTx.reverted ? true : null,
                contractName: contractName || officialContractName,
                counterpartyName: null,
                timestamp,
            };
            if (decodedTx.reverted) {
                interpretation.actions.push("unknown" /* Action.unknown */);
            }
            let fromName = entityMap[fromAddress];
            let toName = entityMap[toAddress || ''];
            if (fromName)
                interpretation.fromName = fromName;
            if (toName)
                interpretation.toName = toName;
            fromName = fromName || fromAddress.substring(0, 6);
            toName = toName || (toAddress || '').substring(0, 6);
            if (interpretation.reverted) {
                interpretation.exampleDescription = 'transaction reverted';
                return interpretation;
                // TODO the description and extras should be different for reverted transactions
            }
            const interpretationMapping = (toAddress && this.contractSpecificInterpreters[toAddress.toLowerCase()]) || null;
            const methodSpecificMappings = [];
            // If multicall, use trace calls to get method speicifc mappings and actions
            // If not, use main method call
            if (constants_1.multicallFunctionNames.includes(methodName || '') && constants_1.multicallContractAddresses.includes(toAddress || '')) {
                traceCalls === null || traceCalls === void 0 ? void 0 : traceCalls.forEach((call) => {
                    const newMethodSpecificMapping = interpretationMapping === null || interpretationMapping === void 0 ? void 0 : interpretationMapping.writeFunctions[call.name || ''];
                    if (newMethodSpecificMapping) {
                        methodSpecificMappings.push(newMethodSpecificMapping);
                    }
                });
            }
            else {
                const newMethodSpecificMapping = interpretationMapping === null || interpretationMapping === void 0 ? void 0 : interpretationMapping.writeFunctions[methodName || ''];
                if (newMethodSpecificMapping) {
                    methodSpecificMappings.push(newMethodSpecificMapping);
                }
            }
            // if there's no contract-specific mapping, try to use the fallback mapping
            if (decodedTx.txType === "contract deploy" /* TxType.CONTRACT_DEPLOY */) {
                // Contract deploy
                interpretation.actions = ["deployed" /* Action.deployed */];
                interpretation.exampleDescription = ContractDeploy_json_1.default.exampleDescription;
                interpretation.extra = Object.assign(Object.assign({}, interpretation.extra), this.useKeywordMap(interactions, ContractDeploy_json_1.default.keywords, '0x_DOESNT_EXIST', userAddress));
                interpretation.exampleDescription = (0, utils_3.fillDescriptionTemplate)(ContractDeploy_json_1.default.exampleDescriptionTemplate, interpretation);
            }
            else if (decodedTx.txType === "native token transfer" /* TxType.TRANSFER */) {
                // Generic transfer
                (0, transfer_1.default)(decodedTx, interpretation, fromName, toName);
            }
            else if (interpretationMapping && methodSpecificMappings.length && methodName && toAddress) {
                // Contract-specific interpretation
                // some of these will be arbitrary keys
                interpretation.contractName = interpretationMapping.contractName;
                for (const methodSpecificMapping of methodSpecificMappings) {
                    if (methodSpecificMapping.action !== "__NFTSALE__" /* Action.__NFTSALE__ */) {
                        interpretation.actions.push(methodSpecificMapping.action);
                    }
                    else {
                        interpretation.actions.push((0, DoubleSidedTxInterpreter_1.getActionForDoubleSidedTx)(interpretation));
                    }
                }
                interpretation.exampleDescription = methodSpecificMappings[0].exampleDescription;
                if (decodedTx.reverted) {
                    interpretation.reverted = true;
                    // TODO the description and extras should be different for reverted transactions
                }
                interpretation.extra = this.useKeywordMap(interactions, methodSpecificMappings[0].keywords, toAddress, userAddress);
                interpretation.exampleDescription = (0, utils_3.fillDescriptionTemplate)(interpretationMapping.writeFunctions[methodName].exampleDescriptionTemplate, interpretation);
            }
            else {
                // TODO remove hardcode on 'execTransaction
                if (decodedTx.contractType === "Gnosis Safe" /* ContractType.GNOSIS */ || decodedTx.methodCall.name === 'execTransaction') {
                    (0, gnosis_1.default)(decodedTx, interpretation);
                }
                else if (decodedTx.contractType !== "OTHER" /* ContractType.OTHER */) {
                    (0, token_1.default)(decodedTx, interpretation);
                }
                else {
                    (0, lastFallback_1.default)(decodedTx, interpretation);
                }
            }
            interpretation.exampleDescription = (0, utils_3.shortenNamesInString)(interpretation.exampleDescription);
            return interpretation;
        });
    }
    getNativeTokenValueSent(interactions, nativeValueSent, fromAddress, userAddress) {
        if (fromAddress === userAddress)
            return Number((0, utils_1.formatEther)(nativeValueSent || 0))
                .toString()
                .replace(/^(\d+\.\d*?[0-9])0+$/g, '$1');
        const nativeTokenEvents = (0, utils_3.getNativeTokenValueEvents)(interactions);
        const nativeTokenEventsReceived = nativeTokenEvents.filter((event) => event.params.from === userAddress);
        const val = nativeTokenEventsReceived.reduce((acc, event) => acc + Number((0, utils_1.formatEther)(event.params.value || 0)), 0);
        return val.toFixed(20).replace(/^(\d+\.\d*?[0-9])0+$/g, '$1');
    }
    getNativeTokenValueReceived(interactions, userAddress) {
        const nativeTokenEvents = (0, utils_3.getNativeTokenValueEvents)(interactions);
        const nativeTokenEventsReceived = nativeTokenEvents.filter((event) => event.params.to === userAddress);
        const val = nativeTokenEventsReceived.reduce((acc, event) => acc + Number((0, utils_1.formatEther)(event.params.value || 0)), 0);
        return val.toFixed(20).replace(/^(\d+\.\d*?[0-9])0+$/g, '$1');
    }
    getWethValue(interactions, userAddress, direction) {
        const eventDirection = direction === 'received' ? 'Deposit' : 'Withdraw';
        const wethInteraction = interactions.find((interaction) => interaction.contractAddress === this.chain.wethAddress);
        if (!wethInteraction)
            return '0';
        const events = wethInteraction.events.filter((event) => event.eventName === eventDirection && event.params.dst === userAddress);
        const val = events.reduce((acc, event) => acc + Number((0, utils_1.formatEther)(event.params.wad || 0)), 0);
        return val.toFixed(20).replace(/^(\d+\.\d*?[0-9])0+$/g, '$1');
    }
    findValue(interactions, keyMapping, userAddress, contractAddress) {
        var _a, _b;
        const filters = keyMapping.filters || {};
        const index = keyMapping.index || 0;
        const array = keyMapping.array || false;
        let decimals = keyMapping.decimals || null;
        let filteredInteractions = deepCopy(interactions);
        for (const [key, value] of Object.entries(filters)) {
            let valueToFind = value;
            // special case for when we want to match on the contextualized user address
            if (value === '{userAddress}')
                valueToFind = userAddress;
            if (value === '{contractAddress}')
                valueToFind = contractAddress;
            // filter out by events that don't have the keys we want
            for (const interaction of filteredInteractions) {
                if (!includes(topLevelInteractionKeys, key)) {
                    interaction.events = interaction.events.filter((d) => (0, utils_3.checkMultipleKeys)(d, key, valueToFind) ||
                        d.params[key] === valueToFind ||
                        d.eventName === valueToFind);
                }
            }
            // if there aren't any events left, we don't want that interaction
            filteredInteractions = filteredInteractions.filter((i) => i.events.length > 0);
            // filter out interactions by top level key
            if (includes(topLevelInteractionKeys, key)) {
                filteredInteractions = filteredInteractions.filter((i) => i[key] === valueToFind);
            }
        }
        let value = null;
        if (!array) {
            const interaction = filteredInteractions[index];
            const prefix = keyMapping.prefix || '';
            let str;
            // unclear if "index" refers specifies an interaction or an event
            // this logic figures that out
            if (interaction === null || interaction === void 0 ? void 0 : interaction[keyMapping.key]) {
                str = interaction === null || interaction === void 0 ? void 0 : interaction[keyMapping.key];
            }
            else {
                const flattenedEvents = (0, utils_3.flattenEventsFromInteractions)(filteredInteractions);
                str =
                    (0, utils_3.checkMultipleKeys)(flattenedEvents[index], keyMapping.key) ||
                        ((_a = flattenedEvents[index]) === null || _a === void 0 ? void 0 : _a.params[keyMapping.key]);
            }
            const postfix = keyMapping.postfix || '';
            value = str ? prefix + str + postfix : keyMapping.defaultValue;
            decimals = decimals || (0, utils_3.getDecimals)(interaction === null || interaction === void 0 ? void 0 : interaction.contractAddress, this.chain);
        }
        else {
            value = [];
            const prefix = keyMapping.prefix || '';
            const postfix = keyMapping.postfix || '';
            // I dont think anything but params will ever use this array feature, so we only support params here for now
            // if the key is in the interaction-level, just loop through those
            // if ((filteredInteractions[0] as any)?.[keyMapping.key]) {
            //     for (const interaction of filteredInteractions) {
            //         const str = (interaction as any)?.[keyMapping.key]
            //         const fullStr = str ? prefix + str + postfix : keyMapping.defaultValue
            //         value.push(fullStr)
            //     }
            // } else {
            for (const interaction of filteredInteractions) {
                for (const event of interaction.events) {
                    const str = event.params[keyMapping.key];
                    const fullStr = str ? prefix + str + postfix : keyMapping.defaultValue;
                    value.push(fullStr);
                }
            }
            //}
            decimals = decimals || (0, utils_3.getDecimals)((_b = filteredInteractions[0]) === null || _b === void 0 ? void 0 : _b.contractAddress, this.chain);
        }
        if (typeof value === 'string' && Number(value)) {
            value = (0, utils_1.formatUnits)(value, decimals);
        }
        return value;
    }
    getTokens(interactions, userAddress, direction) {
        let tokens = [];
        let filteredInteractions = deepCopy(interactions);
        const transferEvents = ['Transfer', 'TransferBatch', 'TransferSingle'];
        for (const interaction of filteredInteractions) {
            // filter non-transfer events
            interaction.events = interaction.events.filter((d) => transferEvents.includes(d.eventName || ''));
            // filter out events that aren't to/from the user in the right direction
            interaction.events = interaction.events.filter((d) => (0, utils_3.toOrFromUser)(d, direction, userAddress));
        }
        // filter out interactions without any of the events we want
        filteredInteractions = filteredInteractions.filter((i) => i.events.filter((d) => transferEvents.includes(d.eventName || '')).length > 0);
        const flattenedInteractions = flattenEvents(filteredInteractions);
        function getTokenType(interaction) {
            const LPTokenSymbols = ['UNI-V2'];
            let tokenType = "unknown" /* AssetType.DEFAULT */;
            // LP Token
            if (interaction.contractSymbol && LPTokenSymbols.includes(interaction.contractSymbol)) {
                tokenType = "LPToken" /* AssetType.LPToken */;
                // ERC-1155
            }
            else if (interaction.contractType == "ERC1155" /* ContractType.ERC1155 */ ||
                interaction.params._amount ||
                interaction.params._amounts) {
                tokenType = "ERC1155" /* AssetType.ERC1155 */;
                // ERC-721
            }
            else if (interaction.contractType == "ERC721" /* ContractType.ERC721 */ || interaction.params.tokenId) {
                tokenType = "ERC721" /* AssetType.ERC721 */;
                // ERC-20
            }
            else if (interaction.contractType == "ERC20" /* ContractType.ERC20 */ || Number(interaction.params.value) > 0) {
                tokenType = "ERC20" /* AssetType.ERC20 */;
            }
            return tokenType;
        }
        // each transfer event can have a token in it. for example, minting multiple NFTs
        function flattenEvents(interactions) {
            const flattenedInteractions = [];
            let flattenedInteraction;
            for (const interaction of interactions) {
                const copiedInteraction = deepCopy(interaction);
                delete copiedInteraction.events;
                flattenedInteraction = copiedInteraction;
                for (const event of interaction.events) {
                    if (event.params._amounts) {
                        // ERC1155 you can batchTransfer multiple tokens, each w their own amount, in 1 event
                        for (const [index, amount] of event.params._amounts.entries()) {
                            let newInteraction = deepCopy(flattenedInteraction);
                            const newEvent = deepCopy(event);
                            newEvent.params._id = event.params._ids
                                ? ethers_1.BigNumber.from(event.params._ids[index]).toString()
                                : null;
                            delete newEvent.params._ids;
                            newEvent.params._amount = ethers_1.BigNumber.from(amount).toString();
                            delete newEvent.params._amounts;
                            newInteraction = Object.assign(Object.assign({}, newInteraction), newEvent);
                            flattenedInteractions.push(newInteraction);
                        }
                    }
                    else {
                        flattenedInteraction = Object.assign(Object.assign({}, flattenedInteraction), event);
                        flattenedInteractions.push(flattenedInteraction);
                    }
                }
            }
            return flattenedInteractions;
        }
        tokens = flattenedInteractions.map((i) => {
            const tokenType = getTokenType(i);
            const amount = i.params.value || i.params.amount || i.params._amount || i.params.wad;
            const tokenId = (i.params.id || i.params.tokenId || '').toString();
            const token = {
                type: tokenType,
                name: i.contractName,
                symbol: i.contractSymbol,
                address: utils_2.AddressZ.parse(i.contractAddress),
            };
            if (amount) {
                let decimals;
                if (tokenType === "ERC1155" /* AssetType.ERC1155 */) {
                    decimals = 0;
                }
                else {
                    decimals = (0, utils_3.getDecimals)(i.contractAddress, this.chain);
                }
                const amountNumber = Number((0, utils_1.formatUnits)(amount, decimals));
                token.amount = amountNumber.toFixed(12).replace(/^(\d+\.\d*?[0-9])0+$/g, '$1');
                // token.amount = amount
            }
            tokenId ? (token.tokenId = tokenId) : null;
            return token;
        });
        return tokens;
    }
    getAssetsReceived(interactions, userAddress) {
        const assets = this.getTokens(interactions, userAddress, 'to');
        const nativeValueReceived = this.getNativeTokenValueReceived(interactions, userAddress);
        const wethValueReceived = this.getWethValue(interactions, userAddress, 'received');
        if (Number(nativeValueReceived)) {
            assets.push({
                type: "native" /* AssetType.native */,
                amount: nativeValueReceived,
                name: 'Ethereum',
                symbol: 'ETH',
                address: constants_1.ethAddress,
            });
        }
        if (Number(wethValueReceived)) {
            assets.push({
                type: "ERC20" /* AssetType.ERC20 */,
                amount: wethValueReceived,
                name: 'Wrapped Ether',
                symbol: 'WETH',
                address: this.chain.wethAddress,
            });
        }
        return assets;
    }
    getAssetsSent(interactions, nativeValueSent, userAddress, fromAddress) {
        const assets = this.getTokens(interactions, userAddress, 'from');
        const ethValueSent = this.getNativeTokenValueSent(interactions, nativeValueSent, fromAddress, userAddress);
        const wethValueSent = this.getWethValue(interactions, userAddress, 'sent');
        if (Number(ethValueSent)) {
            assets.push({
                type: "native" /* AssetType.native */,
                amount: ethValueSent,
                name: 'Ethereum',
                symbol: 'ETH',
                address: constants_1.ethAddress,
            });
        }
        if (Number(wethValueSent)) {
            assets.push({
                type: "ERC20" /* AssetType.ERC20 */,
                amount: wethValueSent,
                name: 'Wrapped Ether',
                symbol: 'WETH',
                address: this.chain.wethAddress,
            });
        }
        return assets;
    }
    useKeywordMap(interactions, keywordsMap, contractAddress, userAddress) {
        const keyValueMap = {};
        const ignoreKeys = ['action', 'contractName', 'exampleDescription'];
        for (const [key, value] of Object.entries(keywordsMap).filter(([key]) => !ignoreKeys.includes(key))) {
            // if the value is a string, we can just use it
            if (typeof value === 'string') {
                keyValueMap[key] = value;
                // some data requires searching for it
            }
            else if (typeof value === 'object') {
                keyValueMap[key] = this.findValue(interactions, value, userAddress, contractAddress);
                if (Array.isArray(keyValueMap[key])) {
                    keyValueMap[`${key}Count`] = keyValueMap[key].length.toString();
                }
            }
        }
        return keyValueMap;
    }
}
exports.default = Interpreter;
//# sourceMappingURL=Interpreter.js.map