"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const providers_1 = require("@ethersproject/providers");
const rawData_1 = require("../interfaces/rawData");
const logging_1 = require("../utils/logging");
class RawDataFetcher {
    constructor(provider, covalent = null) {
        this.formatter = new providers_1.Formatter();
        this.provider = provider;
        this.covalent = covalent;
    }
    static validateTxResponse(unvalidatedTxResponse) {
        return rawData_1.TxResponseZ.parse(unvalidatedTxResponse);
    }
    // TODO use this
    static validateTxReceipt(unvalidatedTxReceipt) {
        return rawData_1.TxReceiptZ.parse(unvalidatedTxReceipt);
    }
    static validateTraceLogs(unvalidatedTraceLogs) {
        return unvalidatedTraceLogs.map((tl) => {
            const data = rawData_1.TraceLogZ.safeParse(tl);
            if (data.success) {
                return data.data;
            }
            else {
                console.log('error in validateTraceLogs', data.error, tl);
                (0, logging_1.logError)({ tx_hash: tl.transactionHash, extra: tl }, data.error);
                return tl;
            }
        });
    }
    getTxHashesByBlockNumber(blockNumber) {
        return __awaiter(this, void 0, void 0, function* () {
            const txHashes = yield this.provider.getBlock(Number(blockNumber)).then((block) => block.transactions);
            return txHashes;
        });
    }
    getTxResponse(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            const unformatted = yield this.provider.getTransaction(txHash);
            const unvalidated = this.formatter.transactionResponse(unformatted);
            const txResponse = RawDataFetcher.validateTxResponse(unvalidated);
            return txResponse;
        });
    }
    getTxReceipt(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            const txReceipt = yield this.provider.getTransactionReceipt(txHash);
            const timestamp = yield this.provider.getBlock(txReceipt.blockNumber).then((block) => block.timestamp);
            const validatedAndFormattedTxReceipt = validateAndFormatTxData(txReceipt, timestamp);
            return validatedAndFormattedTxReceipt;
        });
    }
    getTxTrace(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            // console.log('trace logs:', txHash)
            // logDebug({ txHash }, 'getting trace log')
            const unvalidatedTrace = yield this.provider.send('trace_transaction', [txHash]);
            const traceLogs = RawDataFetcher.validateTraceLogs(unvalidatedTrace);
            // console.log('trace logs', traceLogs)
            return traceLogs;
        });
    }
    // could be parallelized, but each has a different dependency graph
    getTxData(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            let txResponse;
            let txReceipt;
            let txTrace;
            try {
                ;
                [txResponse, txReceipt, txTrace] = yield Promise.all([
                    this.getTxResponse(txHash),
                    this.getTxReceipt(txHash),
                    this.getTxTrace(txHash),
                ]);
            }
            catch (e) {
                console.error('error in getTxData', e);
                throw e;
            }
            return {
                txResponse,
                txReceipt,
                txTrace,
            };
        });
    }
    getTxDataFromS3Tx(tx, timestamp) {
        const txDataStringified = numbersToStrings(tx);
        const txReceipt = validateAndFormatTxData(txDataStringified.transactionReceipt, timestamp);
        const txTrace = RawDataFetcher.validateTraceLogs(txDataStringified.trace);
        const formattedTxResponse = this.formatter.transactionResponse(txDataStringified);
        const txResponse = validateAndFormatTxData(formattedTxResponse);
        return {
            txResponse,
            txReceipt,
            txTrace,
        };
    }
    // TODO should we get the timestamp somehow here too?
    getTxDataWithoutTrace(txHash) {
        return __awaiter(this, void 0, void 0, function* () {
            let txResponse;
            let txReceipt;
            try {
                ;
                [txResponse, txReceipt] = yield Promise.all([this.getTxResponse(txHash), this.getTxReceipt(txHash)]);
            }
            catch (e) {
                console.error('error in getTxData', e);
                throw e;
            }
            return {
                txResponse,
                txReceipt,
                txTrace: [],
            };
        });
    }
    getTxDataWithCovalentByAddress(address, includedInitiatedTxs, includedNotInitiatedTxs, limit) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.covalent) {
                throw new Error('no covalent client');
            }
            const allCovalentTxDataArr = yield this.covalent.getTransactionsFor(address, limit);
            const covalentTxDataArr = allCovalentTxDataArr.filter((tx) => {
                if (includedInitiatedTxs && includedNotInitiatedTxs) {
                    return true;
                }
                else if (includedInitiatedTxs) {
                    return tx.from_address === address; // only transactions initiated by the user, no scam airdrops
                }
                else if (includedNotInitiatedTxs) {
                    return tx.from_address !== address; // only transactions not initiated by the user, just airdrops
                }
                else {
                    return false;
                }
            });
            const rawTxDataArr = yield Promise.all(covalentTxDataArr.map((tx) => __awaiter(this, void 0, void 0, function* () {
                return yield this.getTxData(tx.tx_hash);
            })));
            return {
                rawTxDataArr,
                covalentTxDataArr,
            };
        });
    }
    static getContractAddressesFromRawTxData(rawTxData) {
        const { txReceipt } = rawTxData;
        const addresses = [];
        if (txReceipt.to) {
            addresses.push(txReceipt.to);
        }
        txReceipt.logs.forEach(({ address }) => {
            addresses.push(address);
        });
        return [...new Set(addresses)];
    }
}
exports.default = RawDataFetcher;
// const validateTxHash = (txHash: string): TxHash => {
//     const validTxhash = new RegExp(/^0x[a-fA-F0-9]{64}$/)
//     if (!validTxhash.test(txHash)) {
//         throw new Error(`Invalid txHash: ${txHash}`)
//     }
//     return txHash as TxHash
// }
function numbersToStrings(txData) {
    const receiptWithNumbers = txData.transactionReceipt;
    const receipt = {
        transactionHash: receiptWithNumbers.transactionHash,
        transactionIndex: receiptWithNumbers.transactionIndex,
        blockHash: receiptWithNumbers.blockHash,
        blockNumber: receiptWithNumbers.blockNumber,
        cumulativeGasUsed: receiptWithNumbers.cumulativeGasUsed.toString(),
        gasUsed: receiptWithNumbers.gasUsed.toString(),
        contractAddress: receiptWithNumbers.contractAddress,
        logs: receiptWithNumbers.logs,
        logsBloom: receiptWithNumbers.logsBloom,
        root: receiptWithNumbers.root,
        status: receiptWithNumbers.status,
        to: txData.to,
        from: txData.from,
    };
    const txDataStringified = {
        hash: txData.hash,
        nonce: txData.nonce,
        blockHash: txData.blockHash,
        blockNumber: txData.blockNumber,
        transactionIndex: txData.transactionIndex,
        from: txData.from,
        to: txData.to,
        value: txData.value.toString(),
        gas: txData.gas.toString(),
        gasPrice: txData.gasPrice.toString(),
        input: txData.input,
        transactionReceipt: receipt,
        trace: txData.trace,
    };
    return txDataStringified;
}
function validateAndFormatTxData(txData, timestamp = null) {
    const txResponseFormatted = {};
    const addressKeys = ['from', 'to'];
    for (const [key, val] of Object.entries(txData)) {
        if (addressKeys.includes(key) && val) {
            const address = val.toLowerCase();
            txResponseFormatted[key] = address;
            // in EVMTransaction
        }
        else if (typeof val === 'number') {
            txResponseFormatted[key] = val.toString();
        }
        else if (key !== 'transactionReceipt' && key !== 'trace') {
            txResponseFormatted[key] = val;
        }
    }
    if (timestamp) {
        txResponseFormatted.timestamp = timestamp;
    }
    return txResponseFormatted;
}
// function covalentToRawTxData(rawCovalentData: CovalentTxData): TxReceipt {
//     const data = rawCovalentData
//     const txReceipt: TxReceipt = {
//         transactionHash: data.tx_hash,
//         transactionIndex: data.tx_offset,
//         to: data.to_address,
//         from: data.from_address,
//         blockNumber: data.block_height,
//         gasUsed: BigNumber.from(data.gas_spent),
//         effectiveGasPrice: BigNumber.from(data.gas_price),
//         status: data.successful ? 1 : 0,
//         logs: data.log_events.map((log) => ({
//             blockNumber: log.block_height,
//             transactionIndex: data.tx_offset,
//             // blockHash: data. ,
//             // removed: log. ,
//             address: log.sender_address,
//             data: log.raw_log_data,
//             topics: log.raw_log_topics,
//             transactionHash: log.tx_hash,
//             logIndex: log.log_offset,
//         })),
//         // KEEP but Covalent does not have this
//         // contractAddress: string,
//         // blockHash: string,
//         // confirmations: number,
//         // REMOVE, I dont think we need any of these
//         // logsBloom: string,
//         // root?: string,
//         // cumulativeGasUsed: BigNumber,
//         // byzantium: boolean,
//         // type: number;
//     }
//     return txReceipt
// }
//# sourceMappingURL=RawDataFetcher.js.map