"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const zenLedger_1 = require("../interfaces/zenLedger");
const utils_1 = require("../utils");
class TaxFormatter {
    constructor(walletAddress, walletName, chain = utils_1.chains.ethereum) {
        this.rows = [];
        this.walletAddress = walletAddress;
        this.walletName = walletName;
        this.chain = chain;
    }
    format(activityData, walletAddress = null, walletName = '') {
        this.walletAddress = walletAddress ? walletAddress : this.walletAddress;
        this.walletName = walletName ? walletName : this.walletName;
        activityData = activityData.filter(({ interpretedData }) => interpretedData && !interpretedData.reverted);
        for (const activity of activityData) {
            // console.log('formatting activity', activity)
            this.formatSingleActivity(activity.decodedTx, activity.interpretedData);
        }
        return this.rows;
    }
    formatSingleActivity(decodedTx, interpretedData) {
        const { txHash, contractName, contractAddress, assetsReceived, assetsSent, chainSymbol, reverted, gasPaid, timestamp, userAddress, } = interpretedData;
        const getType = (data) => {
            let rowType = null;
            const action = data.actions[0];
            const isNative = data.assetsReceived.find((asset) => asset.type === "native" /* AssetType.native */) ||
                data.assetsSent.find((asset) => asset.type === "native" /* AssetType.native */);
            switch (action) {
                case "received" /* Action.received */:
                case "claimed" /* Action.claimed */:
                    rowType = zenLedger_1.ZenLedgerRowType.receive;
                    break;
                case "sent" /* Action.sent */:
                    rowType = zenLedger_1.ZenLedgerRowType.send;
                    break;
                case "minted" /* Action.minted */:
                    rowType = zenLedger_1.ZenLedgerRowType.buy;
                    break;
                case "swapped" /* Action.swapped */:
                    rowType = zenLedger_1.ZenLedgerRowType.trade;
                    break;
                case "bought" /* Action.bought */:
                    rowType = isNative ? zenLedger_1.ZenLedgerRowType.trade : zenLedger_1.ZenLedgerRowType.buy;
                    break;
                case "sold" /* Action.sold */:
                    rowType = isNative ? zenLedger_1.ZenLedgerRowType.trade : zenLedger_1.ZenLedgerRowType.sell;
                    break;
                case "got airdropped" /* Action.gotAirdropped */:
                    rowType = zenLedger_1.ZenLedgerRowType.airdrop;
                    break;
                case "revoked" /* Action.revoked */:
                case "approved" /* Action.approved */:
                case "transferred ownership" /* Action.transferredOwnership */:
                case "received ownership" /* Action.receivedOwnership */:
                case "deployed" /* Action.deployed */:
                case "executed" /* Action.executed */:
                case "canceled" /* Action.canceled */:
                    rowType = zenLedger_1.ZenLedgerRowType.fee;
                    break;
                default:
                    rowType = zenLedger_1.ZenLedgerRowType.unknown;
                    break;
            }
            return rowType;
        };
        // type TokenTypeOrNative = AssetType | 'native' | null
        const getAmountAndCurrency = (data, direction) => {
            var _a, _b;
            const assets = direction === 'in' ? data.assetsReceived : data.assetsSent;
            const opposite = direction === 'in' ? data.assetsSent : data.assetsReceived;
            let amount = null;
            let currency = null;
            let type = "unknown" /* AssetType.DEFAULT */;
            if (assets.length > 0) {
                type = assets[0].type;
                if (type === "ERC20" /* AssetType.ERC20 */ || type === "native" /* AssetType.native */) {
                    amount = Number(assets[0].amount);
                    currency = assets[0].symbol;
                }
                else if (type === "ERC721" /* AssetType.ERC721 */) {
                    amount = 1;
                    currency = assets[0].symbol + '-' + ((_a = assets[0].tokenId) === null || _a === void 0 ? void 0 : _a.toString().slice(0, 6));
                }
                else if (type === "ERC1155" /* AssetType.ERC1155 */) {
                    amount = Number(assets[0].amount);
                    currency = assets[0].symbol + '-' + ((_b = assets[0].tokenId) === null || _b === void 0 ? void 0 : _b.toString().slice(0, 6));
                }
                else if (type === "LPToken" /* AssetType.LPToken */) {
                    amount = Number(assets[0].amount);
                    currency = opposite[0].symbol + '-' + opposite[1].symbol;
                }
            }
            return [amount, currency, type];
        };
        const userInitiated = this.walletAddress === userAddress;
        const [receivedAmount, receivedSymbol, receivedType] = getAmountAndCurrency(interpretedData, 'in');
        const [sentAmount, sentSymbol, sentType] = getAmountAndCurrency(interpretedData, 'out');
        const dateTime = new Date((timestamp || 0) * 1000);
        const row = {
            explorerUrl: this.chain.blockExplorerUrl + 'tx/' + txHash,
            userInitiated: userInitiated ? 'true' : 'false',
            Type: getType(interpretedData),
            method: decodedTx.methodCall.name || 'UNKNOWN',
            contract: contractName || 'UNKNOWN',
            'In Amount': receivedAmount,
            'In Currency': receivedSymbol,
            'Out Amount': sentAmount,
            'Out Currency': sentSymbol,
            Timestamp: dateTime.toJSON(),
            'Fee Amount': userInitiated ? Number(gasPaid) : 0,
            'Fee Currency': chainSymbol || 'UNKNOWN',
            'Exchange(optional)': 'Unified ETH Wallet',
            'US Based': 'Yes',
            txHash,
            action: interpretedData.actions[0],
            'received type': receivedType,
            'received amount': receivedAmount,
            'received symbol': receivedSymbol,
            'received type 2': assetsReceived[1] ? assetsReceived[1].type : null,
            'received amount 2': assetsReceived[1] ? Number(assetsReceived[1].amount) : null,
            'received symbol 2': assetsReceived[1] ? assetsReceived[1].symbol : null,
            'sent type': sentType,
            'sent amount': sentAmount,
            'sent symbol': sentSymbol,
            'sent type 2': assetsSent[1] ? assetsSent[1].type : null,
            'sent amount 2': assetsSent[1] ? Number(assetsSent[1].amount) : null,
            'sent symbol 2': assetsSent[1] ? assetsSent[1].symbol : null,
            toAddress: contractAddress,
            // network: interpretedData.chainSymbol || 'UNKNOWN',
            // walletAddress: this.walletAddress,
            // walletName: this.walletName,
        };
        this.rows.push(row);
    }
}
exports.default = TaxFormatter;
//# sourceMappingURL=TaxFormatter.js.map