"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../../utils");
const constants_1 = require("../../utils/constants");
const vars = {
    ERC1155: {
        contractType: "ERC1155" /* ContractType.ERC1155 */,
        type: "ERC1155" /* AssetType.ERC1155 */,
        transfer: 'TransferSingle',
        transferBatch: 'TransferBatch',
        approval: 'ApprovalForAll',
        to: 'to',
        toENS: 'toENS',
        from: 'from',
        fromENS: 'fromENS',
        owner: 'account',
        ownerENS: 'ownerENS',
        operator: 'operator',
        operatorENS: 'operatorENS',
        approved: 'approved',
        approvedENS: 'approvedENS',
        value: 'amount',
    },
    ERC721: {
        contractType: "ERC721" /* ContractType.ERC721 */,
        type: "ERC721" /* AssetType.ERC721 */,
        transfer: 'Transfer',
        transferBatch: null,
        approval: 'ApprovalForAll',
        to: 'to',
        toENS: 'toENS',
        from: 'from',
        fromENS: 'fromENS',
        owner: 'owner',
        ownerENS: 'ownerENS',
        operator: 'operator',
        operatorENS: 'operatorENS',
        approved: 'approved',
        approvedENS: 'approvedENS',
        value: '__N_A__',
    },
    ERC20: {
        contractType: "ERC20" /* ContractType.ERC20 */,
        type: "ERC20" /* AssetType.ERC20 */,
        transfer: 'Transfer',
        transferBatch: null,
        approval: 'Approval',
        to: 'to',
        toENS: 'toENS',
        from: 'from',
        fromENS: 'fromENS',
        owner: 'owner',
        ownerENS: 'ownerENS',
        operator: 'spender',
        operatorENS: 'spenderENS',
        approved: 'approved',
        approvedENS: 'approvedENS',
        value: 'value',
    },
    WETH: {
        contractType: "WETH" /* ContractType.WETH */,
        type: "ERC20" /* AssetType.ERC20 */,
        transfer: 'Transfer',
        transferBatch: null,
        approval: 'Approval',
        to: 'dst',
        toENS: 'dstENS',
        from: 'src',
        fromENS: 'srcENS',
        owner: 'src',
        ownerENS: 'srcENS',
        operator: 'guy',
        operatorENS: 'guyENS',
        approved: 'approved',
        approvedENS: 'approvedENS',
        value: 'wad',
    },
};
function isMintEvent(t, event, userAddress, fromAddress) {
    return (event.eventName === t.transfer &&
        event.params[t.from] === constants_1.blackholeAddress &&
        event.params[t.to] === userAddress &&
        userAddress === fromAddress // user needs to have initiated the mint
    );
}
function isAirdropEvent(t, event, userAddress, fromAddress) {
    return ((event.eventName === t.transfer || event.eventName === t.transferBatch) &&
        event.params[t.from] === constants_1.blackholeAddress &&
        event.params[t.to] === userAddress &&
        userAddress !== fromAddress // user needs to have initiated the mint
    );
}
function isSendEvent(t, event, userAddress) {
    return ((event.eventName === t.transfer || event.eventName === t.transferBatch) && event.params[t.from] === userAddress);
}
function isReceiveEvent(t, event, userAddress, fromAddress) {
    return event.eventName === t.transfer && event.params[t.to] === userAddress && fromAddress !== userAddress;
}
// function isSwapEvent(t: TokenVars, event: InteractionEvent, userAddress: string) {
//     return (
//         event.eventName === t.transfer &&
//         event.params[t.from] === userAddress &&
//         blackholeAddresses.includes(event.params[t.to] as string)
//     )
// }
function isClaimEvent(t, event, userAddress, fromAddress) {
    return event.eventName === t.transfer && event.params[t.to] === userAddress && fromAddress === userAddress;
}
function isApprovalEvent(t, event, userAddress) {
    return (event.eventName === t.approval &&
        event.params[t.owner] === userAddress &&
        (event.params[t.approved] === true || Number(event.params[t.value]) > 0));
}
function isRevokeEvent(t, event, userAddress) {
    return (event.eventName === t.approval &&
        event.params[t.owner] === userAddress &&
        (event.params[t.approved] === false || Number(event.params[t.value]) === 0));
}
function isBurnEvent(t, event, userAddress) {
    return (event.eventName === t.transfer &&
        event.params[t.from] === userAddress &&
        constants_1.blackholeAddresses.includes(event.params[t.to]));
}
function getAction(t, events, userAddress, fromAddress) {
    const isMint = events.find((e) => isMintEvent(t, e, userAddress, fromAddress));
    const isAirdrop = events.find((e) => isAirdropEvent(t, e, userAddress, fromAddress));
    const isSend = events.find((e) => isSendEvent(t, e, userAddress));
    const isReceive = events.find((e) => isReceiveEvent(t, e, userAddress, fromAddress));
    // const isSwap = events.find((e) => isSwapEvent(t, e, userAddress, fromAddress))
    const isClaim = events.find((e) => isClaimEvent(t, e, userAddress, fromAddress));
    const isApproved = events.find((e) => isApprovalEvent(t, e, userAddress));
    const isRevoked = events.find((e) => isRevokeEvent(t, e, userAddress));
    const isBurn = events.find((e) => isBurnEvent(t, e, userAddress));
    if (isMint) {
        return "minted" /* Action.minted */;
    }
    else if (isBurn) {
        return "burned" /* Action.burned */;
    }
    else if (isAirdrop) {
        return "got airdropped" /* Action.gotAirdropped */;
    }
    else if (isSend) {
        return "sent" /* Action.sent */;
    }
    else if (isReceive) {
        return "received" /* Action.received */;
    }
    else if (isClaim) {
        return "claimed" /* Action.claimed */;
    }
    else if (isApproved) {
        return "approved" /* Action.approved */;
    }
    else if (isRevoked) {
        return "revoked" /* Action.revoked */;
    }
    return "______TODO______" /* Action.______TODO______ */;
}
function getTokenInfo(tokenContractInteraction, interpretation) {
    const { actions, assetsReceived, assetsSent } = interpretation;
    if (actions.length) {
        switch (actions[0]) {
            case "minted" /* Action.minted */:
            case "got airdropped" /* Action.gotAirdropped */:
            case "received" /* Action.received */:
            case "claimed" /* Action.claimed */:
                return assetsReceived[0];
            case "sent" /* Action.sent */:
            case "burned" /* Action.burned */:
                return assetsSent[0];
            case "approved" /* Action.approved */:
            case "revoked" /* Action.revoked */:
                return {
                    type: "ERC721" /* AssetType.ERC721 */,
                    name: tokenContractInteraction === null || tokenContractInteraction === void 0 ? void 0 : tokenContractInteraction.contractName,
                    symbol: tokenContractInteraction === null || tokenContractInteraction === void 0 ? void 0 : tokenContractInteraction.contractSymbol,
                    address: tokenContractInteraction === null || tokenContractInteraction === void 0 ? void 0 : tokenContractInteraction.contractAddress,
                };
            default:
                // console.log('getTokenInfo action not supported: ')
                return {
                    type: "unknown" /* AssetType.DEFAULT */,
                    name: '',
                    symbol: '',
                    address: '0x',
                };
        }
    }
    return {
        type: "unknown" /* AssetType.DEFAULT */,
        name: '',
        symbol: '',
        address: '0x',
    };
}
function addUserName(t, interpretation, tokenEvents, userAddress, fromAddress) {
    let userName = interpretation.userName;
    if (interpretation.actions.length) {
        switch (interpretation.actions[0]) {
            case 'received':
                userName = tokenEvents
                    .filter((e) => isReceiveEvent(t, e, userAddress, fromAddress))
                    .map((e) => e.params[t.toENS] || e.params[t.to])[0];
                break;
            default:
                break;
        }
    }
    interpretation.userName = (0, utils_1.shortenName)(userName);
}
function addCounterpartyNames(t, interpretation, tokenEvents, userAddress, fromAddress) {
    let counterpartyNames = [];
    if (interpretation.actions.length) {
        switch (interpretation.actions[0]) {
            case 'received':
                counterpartyNames = tokenEvents
                    .filter((e) => isReceiveEvent(t, e, userAddress, fromAddress))
                    .map((e) => e.params[t.fromENS] || e.params[t.from]);
                break;
            case 'sent':
                counterpartyNames = tokenEvents
                    .filter((e) => isSendEvent(t, e, userAddress))
                    .map((e) => e.params[t.toENS] || e.params[t.to]);
                break;
            case 'approved':
                counterpartyNames = tokenEvents
                    .filter((e) => isApprovalEvent(t, e, userAddress))
                    .map((e) => e.params[t.operatorENS] || e.params[t.operator]);
                break;
            case 'revoked':
                counterpartyNames = tokenEvents
                    .filter((e) => isRevokeEvent(t, e, userAddress))
                    .map((e) => e.params[t.operatorENS] || e.params[t.operator]);
                break;
            case 'got airdropped':
                counterpartyNames = [fromAddress];
                break;
            default:
                break;
        }
    }
    counterpartyNames = counterpartyNames.map((n) => (0, utils_1.shortenName)(n));
    if (counterpartyNames.length === 1) {
        interpretation.counterpartyName = counterpartyNames[0];
    }
    if (counterpartyNames.length > 1) {
        interpretation.extra.counterpartyNames = counterpartyNames;
    }
}
function addExampleDescription(interpretation, token) {
    var _a;
    const { assetsReceived: tokensReceived, assetsSent: tokensSent } = interpretation;
    let exampleDescription = '';
    const i = interpretation;
    const { userName, actions: action } = i;
    const { name: tokenName } = token;
    const receivedSingle = tokensReceived.length === 1;
    const sentSingle = tokensSent.length === 1;
    const symbol = token.symbol ? ' ' + token.symbol : ' ???';
    let tokenId = token.tokenId ? ' #' + token.tokenId : '';
    const amount = token.amount ? ' ' + token.amount : '';
    let tokenCount = 0;
    let counterpartyName = ((_a = i.extra.counterpartyNames) === null || _a === void 0 ? void 0 : _a.length) ? i.extra.counterpartyNames[0] : i.counterpartyName || '';
    counterpartyName = counterpartyName ? ' ' + counterpartyName : '';
    let direction = '';
    if (interpretation.actions.length) {
        switch (action[0]) {
            case "minted" /* Action.minted */: {
                tokenId = receivedSingle ? tokenId : '';
                tokenCount = i.assetsReceived.filter((t) => t.address === token.address).length;
                counterpartyName = ' ' + tokenName;
                direction = ' from';
                break;
            }
            case "received" /* Action.received */:
            case "got airdropped" /* Action.gotAirdropped */:
            case "revoked" /* Action.revoked */:
                tokenId = receivedSingle ? tokenId : '';
                tokenCount = i.assetsReceived.length;
                direction = ' from';
                break;
            case "sent" /* Action.sent */:
                tokenId = sentSingle ? tokenId : '';
                tokenCount = i.assetsSent.length;
                direction = ' to';
                break;
            case "approved" /* Action.approved */:
                direction = ' to be managed by';
                break;
            default:
                break;
        }
    }
    tokenCount = tokenCount > 1 ? ' ' + tokenCount : '';
    exampleDescription = `${userName} ${action}${amount}${tokenCount}${symbol}${tokenId}${direction}${counterpartyName}`;
    interpretation.exampleDescription = exampleDescription;
}
function interpretGenericToken(decodedData, interpretation) {
    const { userAddress } = interpretation;
    const { fromAddress, toAddress, interactions, contractType, contractName } = decodedData;
    if (contractType === "OTHER" /* ContractType.OTHER */ || contractType === "Gnosis Safe" /* ContractType.GNOSIS */) {
        throw new Error('Token type not supported');
    }
    const t = vars[contractType];
    const defaultTokenInteraction = {
        contractName,
        contractSymbol: null,
        contractAddress: toAddress || '',
        contractType: "OTHER" /* ContractType.OTHER */,
        events: [],
    };
    const tokenContractInteraction = interactions.find((interaction) => interaction.contractAddress === toAddress) || defaultTokenInteraction;
    const tokenEvents = (tokenContractInteraction === null || tokenContractInteraction === void 0 ? void 0 : tokenContractInteraction.events) || [];
    // use TokensReceived and TokensSent to determine the actions (besides mint), not the events. Bought, sold, traded
    interpretation.actions.push(getAction(t, tokenEvents, userAddress, fromAddress));
    const token = getTokenInfo(tokenContractInteraction, interpretation);
    addUserName(t, interpretation, tokenEvents, userAddress, fromAddress);
    addCounterpartyNames(t, interpretation, tokenEvents, userAddress, fromAddress);
    addExampleDescription(interpretation, token);
}
exports.default = interpretGenericToken;
//# sourceMappingURL=token.js.map