"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("ethers/lib/utils");
const ethersAbiCoder = new utils_1.AbiCoder(function (type, value) {
    if (type.match(/^u?int/) &&
        !Array.isArray(value) &&
        (!(!!value && typeof value === 'object') || value.constructor.name !== 'BigNumber')) {
        return value.toString();
    }
    return value;
});
class ABICoder {
    mapTypes(types) {
        const mappedTypes = [];
        types.forEach((type) => {
            // Remap `function` type params to bytes24 since Ethers does not
            // recognize former type. Solidity docs say `Function` is a bytes24
            // encoding the contract address followed by the function selector hash.
            if (typeof type === 'object' && type.type === 'function') {
                type = Object.assign({}, type, { type: 'bytes24' });
            }
            if (this.isSimplifiedStructFormat(type)) {
                const structName = Object.keys(type)[0];
                mappedTypes.push(Object.assign(this.mapStructNameAndType(structName), {
                    components: this.mapStructToCoderFormat(type[structName]),
                }));
                return;
            }
            mappedTypes.push(type);
        });
        return mappedTypes;
    }
    isSimplifiedStructFormat(type) {
        return typeof type === 'object' && typeof type.components === 'undefined' && typeof type.name === 'undefined';
    }
    mapStructNameAndType(structName) {
        let type = 'tuple';
        if (structName.indexOf('[]') > -1) {
            type = 'tuple[]';
            structName = structName.slice(0, -2);
        }
        return { type: type, name: structName };
    }
    mapStructToCoderFormat(struct) {
        const components = [];
        Object.keys(struct).forEach((key) => {
            if (typeof struct[key] === 'object') {
                components.push(Object.assign(this.mapStructNameAndType(key), {
                    components: this.mapStructToCoderFormat(struct[key]),
                }));
                return;
            }
            components.push({
                name: key,
                type: struct[key],
            });
        });
        return components;
    }
    decodeParameters(outputs, bytes) {
        if (outputs.length > 0 && (!bytes || bytes === '0x' || bytes === '0X')) {
            throw new Error("Returned values aren't valid, did it run Out of Gas? " +
                'You might also see this error if you are not using the ' +
                'correct ABI for the contract you are retrieving data from, ' +
                'requesting data from a block number that does not exist, ' +
                'or querying a node which is not fully synced.');
        }
        const res = ethersAbiCoder.decode(this.mapTypes(outputs), '0x' + bytes.replace(/0x/i, ''), false);
        const errors = (0, utils_1.checkResultErrors)(res);
        if (errors.length)
            throw errors[0].error;
        const returnValue = {};
        outputs.forEach((output, i) => {
            let decodedValue = res[Object.keys(returnValue).length];
            const isStringObject = typeof output === 'object' && output.type && output.type === 'string';
            const isStringType = typeof output === 'string' && output === 'string';
            // only convert `0x` to null if it's not string value
            decodedValue = decodedValue === '0x' && !isStringObject && !isStringType ? null : decodedValue;
            returnValue[i] = decodedValue;
            // WARNING: don't think we need this for our uses, it was causing the object to have more keys than we expected, so we commented it out.
            // if ((typeof output === 'function' || (!!output && typeof output === 'object')) && output.name) {
            //     returnValue[output.name] = decodedValue
            // }
        });
        return returnValue;
    }
}
exports.default = ABICoder;
//# sourceMappingURL=web3-abi-coder.js.map